from rest_framework.schemas.openapi import (
    SchemaGenerator as DRFSchemaGenerator,
    AutoSchema as DRFAutoSchema,
)
from camomilla.contrib.rest_framework.serializer import (
    TranslationsMixin,
    plain_to_nest,
    TRANS_ACCESSOR,
)


class AutoSchema(DRFAutoSchema):
    def map_serializer(self, serializer):
        schema = super(AutoSchema, self).map_serializer(serializer)
        if isinstance(serializer, TranslationsMixin) and serializer.is_translatable:
            schema = plain_to_nest(schema["properties"], serializer.translation_fields)
            schema[TRANS_ACCESSOR] = {
                "type": "object",
                "properties": {
                    k: {"type": "object", "properties": v}
                    for k, v in schema[TRANS_ACCESSOR].items()
                },
            }
        return schema


class SchemaGenerator(DRFSchemaGenerator):
    def create_view(self, callback, method, request=None):
        view = super(SchemaGenerator, self).create_view(callback, method, request)
        view.schema = AutoSchema()
        return view



import traceback
from io import BytesIO

from django.core.files.base import ContentFile
from django.core.files.storage import get_storage_class
from PIL import Image

from camomilla import settings


class OptimizedStorage(get_storage_class()):
    MEDIA_MAX_WIDTH = settings.MEDIA_OPTIMIZE_MAX_WIDTH
    MEDIA_MAX_HEIGHT = settings.MEDIA_OPTIMIZE_MAX_HEIGHT
    MEDIA_DPI = settings.MEDIA_OPTIMIZE_DPI

    def __init__(self, *args, **kwargs) -> None:
        self.MEDIA_MAX_WIDTH = kwargs.pop("max_width", self.MEDIA_MAX_WIDTH)
        self.MEDIA_MAX_HEIGHT = kwargs.pop("max_height", self.MEDIA_MAX_HEIGHT)
        self.MEDIA_DPI = kwargs.pop("dpi", self.MEDIA_DPI)
        super().__init__(*args, **kwargs)

    def _save(self, name, content):
        if settings.ENABLE_MEDIA_OPTIMIZATION:
            content, _ = self._optimize(name, content)
        return super(OptimizedStorage, self)._save(name, content)

    def _optimize(self, name, content):
        try:
            image = Image.open(content)
            width, height = image.size
            if width <= height:
                selected_width = int((self.MEDIA_MAX_HEIGHT / height) * width)
                selected_height = self.MEDIA_MAX_HEIGHT
            else:
                selected_height = int((self.MEDIA_MAX_WIDTH / width) * height)
                selected_width = self.MEDIA_MAX_WIDTH

            image = image.resize(
                [selected_width, selected_height], resample=Image.LANCZOS
            )
            image.info["dpi"] = (self.MEDIA_DPI, self.MEDIA_DPI)
            tmp = BytesIO()
            ext = name.split(".")[-1].lower().replace("jpg", "jpeg")
            image.save(tmp, ext, dpi=(self.MEDIA_DPI, self.MEDIA_DPI), optimize=True)
            tmp.seek(0)
            optimized_content = ContentFile(tmp.read())
            tmp.close()
            content.close()
            return optimized_content, True
        except:
            traceback.print_exc()
            return content, False



def build_standard_model_serializer(model, depth, bases=None):
    if bases is None:
        from rest_framework.serializers import ModelSerializer
        from camomilla.serializers.fields import FieldsOverrideMixin
        from camomilla.serializers.mixins import (
            JSONFieldPatchMixin,
            NestMixin,
            OrderingMixin,
            SetupEagerLoadingMixin,
        )

        bases = (
            NestMixin,
            FieldsOverrideMixin,
            JSONFieldPatchMixin,
            OrderingMixin,
            SetupEagerLoadingMixin,
            ModelSerializer,
        )
    return type(
        f"{model.__name__}StandardSerializer",
        bases,
        {
            "Meta": type(
                "Meta",
                (object,),
                {"model": model, "depth": depth, "fields": "__all__"},
            )
        },
    )



from collections import defaultdict
from typing import Any

from django.db.models import JSONField
from django.db.models.query_utils import DeferredAttribute

from .fields import *
from .models import *
from .models import _Cache, build_model_cache


class StructuredDescriptior(DeferredAttribute):
    def __set__(self, instance, value):
        if isinstance(value, dict):
            self.field.prefetch_related(value)
            value = self.field.populate_schema(value)
        elif isinstance(value, list):
            self.field.prefetch_related(value)
            _stack = []
            for v in value:
                if isinstance(v, dict):
                    v = self.field.populate_schema(v)
                elif not isinstance(v, self.field.schema):
                    raise TypeError(
                        f"{type(v)} is not a valid type for the given schema ({self.field.schema})."
                    )
                v.validate()
                _stack.append(v)
            value = _stack
        elif isinstance(value, self.field.schema):
            value.validate()
        else:
            raise TypeError(
                f"{type(value)} is not a valid type for the given schema ({self.field.schema})."
            )
        instance.__dict__[self.field.attname] = value


class StructuredJSONField(JSONField):
    # TODO: share cache in querysets of models having this same field
    # TODO: write queries for prefetch related for models inside the field

    descriptor_class = StructuredDescriptior

    def __init__(self, schema, *args: Any, **kwargs: Any) -> None:
        self.schema = schema
        self._cache = _Cache(self)
        return super().__init__(*args, **kwargs)

    def populate_schema(self, struct):
        schema_intance = self.schema()
        relations = schema_intance.prepopulate(**struct)
        schema_intance.bind(self)
        schema_intance.populate(**relations)
        return schema_intance

    def get_prep_value(self, value):
        if not value:
            return super().get_prep_value(value)
        if isinstance(value, list):
            value = [
                (v if isinstance(v, self.schema) else self.populate_schema(v))
                for v in value
            ]
            value = [self.schema.to_db_transform(v.to_struct()) for v in value]
        elif isinstance(value, dict):
            value = self.schema.to_db_transform(self.populate_schema(value).to_struct())
        elif isinstance(value, self.schema):
            value = self.schema.to_db_transform(value.to_struct())
        else:
            raise TypeError(f"{type(value)} is not a valid type for the given schema.")
        return super().get_prep_value(value)

    def from_db_value(self, value, expression, connection):
        return self.schema.from_db_transform(
            super().from_db_value(value, expression, connection)
        )

    def get_prefetched_data(self):
        return self._cache.get_prefetched_data()

    def get_all_relateds(self, struct):
        if isinstance(struct, list):
            relateds = defaultdict(set)
            for inner_struct in struct:
                if isinstance(inner_struct, dict):
                    child_relateds = self.schema.get_all_relateds(inner_struct)
                    for model, pks in child_relateds.items():
                        relateds[model].update(pks)
            return relateds
        return self.schema.get_all_relateds(struct)

    def prefetch_related(self, struct):
        relateds = self.get_all_relateds(struct)
        for model, pks in relateds.items():
            self._cache.prefetched_data[model] = (
                {obj.pk: obj for obj in build_model_cache(model, pks)}
                if len(pks) > 0
                else {}
            )

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        kwargs["schema"] = self.schema
        return name, path, args, kwargs



import datetime
import decimal

from django.core import validators
from django.db import models
from django.utils.dateparse import parse_duration
from django.utils.duration import duration_string
from jsonmodels import fields
from jsonmodels.validators import ValidationError
from camomilla.utils.getters import pointed_getter


__all__ = [
    "CharField",
    "DecimalField",
    "IntegerField",
    "FloatField",
    "BooleanField",
    "DateTimeField",
    "DateField",
    "TimeField",
    "DurationField",
    "EmailField",
    "SlugField",
    "URLField",
    "ListField",
    "DictField",
    "EmbeddedField",
    "ForeignKey",
    "ForeignKeyList",
]


class Field(fields.BaseField):
    additional_kwargs = {}
    additional_validators = []
    __validators = []
    parent = None

    def __init__(self, *args, **kwargs):
        for key, value in self.additional_kwargs.items():
            setattr(self, key, kwargs.pop(key, value))
        super(Field, self).__init__(*args, **kwargs)

    @property
    def validators(self):
        validators = self.__validators
        for validator in self.additional_validators:
            if isinstance(validator, dict):
                validator_class = validator["validator"]
                args = [
                    getattr(self, arg)
                    for arg in validator.get("args", [])
                    if getattr(self, arg, fields.NotSet) != fields.NotSet
                ]
                kwargs = {
                    key: getattr(self, value)
                    for key, value in validator.get("kwargs", {}).items()
                    if getattr(self, value, fields.NotSet) != fields.NotSet
                }
                if len(args) or len(kwargs.keys()):
                    validators.append(validator_class(*args, **kwargs))
            else:
                validators.append(validator)
        return validators

    @validators.setter
    def validators(self, value):
        self.__validators = value

    def bind(self, parent):
        self.parent = parent
    
    @classmethod
    def to_db_transform(cls, data):
        return data

    @classmethod
    def from_db_transform(cls, data):
        return data


class CharField(fields.StringField, Field):
    additional_kwargs = {
        "max_length": fields.NotSet,
    }
    additional_validators = [
        {"validator": validators.MaxLengthValidator, "args": ["max_length"]}
    ]


class DecimalField(fields.StringField, Field):
    types = (decimal.Decimal, int, float)

    additional_kwargs = {
        "max_digits": fields.NotSet,
        "decimal_places": fields.NotSet,
    }
    additional_validators = [
        {
            "validator": validators.DecimalValidator,
            "args": ["max_digits", "decimal_places"],
        }
    ]


class IntegerField(fields.IntField, Field):
    additional_kwargs = {
        "min_value": fields.NotSet,
        "max_value": fields.NotSet,
    }
    additional_validators = [
        {"validator": validators.MinValueValidator, "args": ["min_value"]},
        {"validator": validators.MaxValueValidator, "args": ["max_value"]},
    ]


class FloatField(IntegerField):
    types = (float, int)


class BooleanField(fields.BoolField, Field):
    pass


class DateTimeField(fields.DateTimeField, Field):
    pass


class DateField(fields.DateField, Field):
    pass


class TimeField(fields.TimeField, Field):
    pass


class DurationField(fields.StringField, Field):
    types = (datetime.timedelta,)

    def to_struct(self, value):
        return value and duration_string(value)

    def parse_value(self, value):
        if value is None:
            return value
        if isinstance(value, datetime.timedelta):
            return value
        try:
            parsed = parse_duration(value)
        except ValueError:
            pass
        else:
            if parsed is not None:
                return parsed

        raise fields.ValidationError(
            "“%(value)s” value has an invalid format. It must be in '[DD] [[HH:]MM:]ss[.uuuuuu] format.'".format(
                value=value
            )
        )


class EmailField(CharField):
    additional_validators = [validators.validate_email]


class SlugField(CharField):
    additional_validators = [validators.validate_slug]


class URLField(CharField):
    additional_validators = [validators.URLValidator()]


class ListField(fields.ListField, Field):
    
    def parse_value(self, values):
        """Cast value to proper collection."""
        result = self.get_default_value()
        if not values:
            return result
        if not isinstance(values, list):
            return values
        parent = self.parent
        return [self._cast_value(value, parent) for value in values]
    
    def _cast_value(self, value, parent):
        if isinstance(value, self.items_types):
            return value
        else:
            main_type = self._get_main_type()
            from camomilla.structured.models import Model
            if issubclass(main_type, Model):
                instance = main_type()
                relations = instance.prepopulate(**value)
                instance.bind(parent)
                instance.populate(**relations)
                return instance
            return main_type(**value)

    def _get_main_type(self):
        if len(self.items_types) != 1:
            raise ValidationError(
                'Cannot decide which type to choose from "{items_types}".'.format(
                    items_types=", ".join([t.__name__ for t in self.items_types])
                )
            )
        return self.items_types[0]
    
class DictField(fields.DictField, Field):
    pass


class EmbeddedField(fields.EmbeddedField, Field):
    def parse_value(self, value):
        if not isinstance(value, dict):
            return value
        embed_instance = self._get_embed_type()()
        relations = embed_instance.prepopulate(**value)
        embed_instance.bind(self.parent)
        embed_instance.populate(**relations)
        return embed_instance


class ForeignKey(Field):
    types = (models.Model, int, str)

    def __init__(self, to, *args, **kwargs):
        self.model = to
        self.queryset = kwargs.pop("queryset", None)
        super(ForeignKey, self).__init__(*args, **kwargs)

    def get_queryset(self):
        return self.queryset or self.model.objects.all()

    def to_struct(self, value):
        return getattr(value, "pk", value)

    def parse_value(self, value):
        try:
            if isinstance(value, bool):
                raise TypeError
            if isinstance(value, self.model):
                return value
            cache = pointed_getter(self, "parent.get_prefetched_data", lambda: None)()
            if cache is not None:
                return cache.get(self.model, {}).get(value)
            return value and self.get_queryset().get(pk=value)
        except (TypeError, ValueError):
            raise fields.ValidationError(
                "Incorrect type. Expected pk value, received {data_type}.".format(
                    data_type=type(value).__name__
                )
            )


class ForeignKeyList(ListField):
    types = (list, models.QuerySet)

    def __init__(self, to, *args, **kwargs):
        self.inner_model = to
        self.inner_queryset = kwargs.pop("queryset", None)
        kwargs["items_types"] = to
        super(ForeignKeyList, self).__init__(*args, **kwargs)

    def parse_value(self, values):
        result = self.get_default_value()
        if not values:
            return result
        if not isinstance(values, list):
            return values
        cache = pointed_getter(self, "parent.get_prefetched_data", lambda: None)()
        if cache is not None:
            data = cache.get(self.inner_model)
            qs = self.get_inner_queryset().filter(pk__in=values)
            setattr(
                qs,
                "_result_cache",
                [data[i] for i in values if i in data] if data else [],
            )
            return qs
        preserved = models.Case(
            *[models.When(pk=pk, then=pos) for pos, pk in enumerate(values)]
        )
        return self.get_inner_queryset().filter(pk__in=values).order_by(preserved)

    def get_inner_queryset(self):
        return self.inner_queryset or self.inner_model._default_manager.all()

    def to_struct(self, value):
        return [getattr(v, "pk", v) for v in value]



import django
from django.conf import settings as django_settings
from django.db.models.aggregates import Max
from django.db.models.functions import Coalesce
from django.utils import translation
from modeltranslation.settings import AVAILABLE_LANGUAGES
from modeltranslation.utils import build_localized_fieldname
from rest_framework import serializers
from rest_framework.utils import model_meta

from camomilla.fields import ORDERING_ACCEPTED_FIELDS
from camomilla.serializers.fields.related import RelatedField
from camomilla.serializers.utils import build_standard_model_serializer
from camomilla.serializers.validators import UniquePermalinkValidator
from camomilla.utils import dict_merge
from camomilla import settings

if django.VERSION >= (4, 0):
    from django.db.models import JSONField as DjangoJSONField
else:
    from django.contrib.postgres.fields import JSONField as DjangoJSONField


# TODO: decide what to do with LangInfoMixin mixin!
class LangInfoMixin(metaclass=serializers.SerializerMetaclass):
    lang_info = serializers.SerializerMethodField("get_lang_info", read_only=True)

    def get_lang_info(self, obj, *args, **kwargs):
        languages = []
        for key, language in django_settings.LANGUAGES:
            languages.append({"id": key, "name": language})
        return {
            "default": django_settings.LANGUAGE_CODE,
            "active": translation.get_language(),
            "site_languages": languages,
        }

    def get_default_field_names(self, *args):
        field_names = super().get_default_field_names(*args)
        self.action = getattr(
            self, "action", self.context and self.context.get("action", "list")
        )
        if self.action != "retrieve":
            return [f for f in field_names if f != "lang_info"]
        return field_names


class SetupEagerLoadingMixin:
    @staticmethod
    def setup_eager_loading(queryset):
        return queryset


class OrderingMixin:
    def get_max_order(self, order_field):
        return self.Meta.model.objects.aggregate(
            max_order=Coalesce(Max(order_field), 0)
        )["max_order"]

    def _get_ordering_field_name(self):
        try:
            field_name = self.Meta.model._meta.ordering[0]
            if field_name[0] == "-":
                field_name = field_name[1:]
            return field_name
        except (AttributeError, IndexError):
            return None

    def build_standard_field(self, field_name, model_field):
        field_class, field_kwargs = super().build_standard_field(
            field_name, model_field
        )
        if (
            isinstance(model_field, ORDERING_ACCEPTED_FIELDS)
            and field_name == self._get_ordering_field_name()
        ):
            field_kwargs["default"] = self.get_max_order(field_name) + 1
        return field_class, field_kwargs


class JSONFieldPatchMixin:
    def is_json_field(self, attr, value, info):
        return (
            attr in info.fields
            and isinstance(info.fields[attr], DjangoJSONField)
            and isinstance(value, dict)
        )

    def update(self, instance, validated_data):
        if self.partial:
            info = model_meta.get_field_info(instance)
            for attr, value in validated_data.items():
                if self.is_json_field(attr, value, info):
                    validated_data[attr] = dict_merge(
                        getattr(instance, attr, {}), value
                    )
        return super().update(instance, validated_data)


class NestMixin:
    def __init__(self, *args, **kwargs):
        self._depth = kwargs.pop("depth", None)
        return super().__init__(*args, **kwargs)

    def build_nested_field(self, field_name, relation_info, nested_depth):
        return self.build_relational_field(field_name, relation_info, nested_depth)

    def build_relational_field(
        self, field_name, relation_info, nested_depth=settings.API_NESTING_DEPTH + 1
    ):
        nested_depth = nested_depth if self._depth is None else self._depth
        field_class, field_kwargs = super().build_relational_field(
            field_name, relation_info
        )
        if (
            field_class is RelatedField and nested_depth > 1
        ):  # stop recursion one step before the jump :P
            field_kwargs["serializer"] = build_standard_model_serializer(
                relation_info[1], nested_depth - 1
            )
        return field_class, field_kwargs



class AbstractPageMixin(serializers.ModelSerializer):
    LANG_PERMALINK_FIELDS = [
        build_localized_fieldname("permalink", lang)
        for lang in AVAILABLE_LANGUAGES
        if settings.ENABLE_TRANSLATIONS
    ]

    @property
    def translation_fields(self):
        return super().translation_fields + ["permalink"]

    def get_default_field_names(self, *args):
        return (
            [f for f in super().get_default_field_names(*args) if f != "url_node"]
            + self.LANG_PERMALINK_FIELDS
            + ["permalink"]
        )

    def build_field(self, field_name, info, model_class, nested_depth):
        if field_name in self.LANG_PERMALINK_FIELDS + ["permalink"]:
            return serializers.CharField, {
                "source": "url_node.%s" % field_name,
                "read_only": True,
            }
        return super().build_field(field_name, info, model_class, nested_depth)

    def get_validators(self):
        return super().get_validators() + [UniquePermalinkValidator()]
 


from collections import defaultdict
from jsonmodels.models import Base
from jsonmodels.fields import _LazyType
from jsonmodels.validators import ValidationError
from camomilla.structured.fields import (
    ForeignKey,
    ForeignKeyList,
    EmbeddedField,
    ListField,
)
from camomilla.utils.getters import pointed_getter

__all__ = ["Model"]


class _Cache:
    instance = None
    parent = None
    prefetched_data = {}

    def __init__(self, instance):
        self.instance = instance
        self.prefetched_data = {}

    @property
    def is_root(self):
        return self.parent == None

    @property
    def root_node(self):
        root = self.instance
        while root._cache.parent is not None:
            root = root._cache.parent
        return root

    def get_prefetched_data(self):
        if self.is_root:
            return self.prefetched_data
        return self.root_node.get_prefetched_data()


class Model(Base):
    def __init__(self, **kwargs):
        self._cache = _Cache(self)
        for _, field in self.iterate_over_fields():
            field.bind(self)
        super(Model, self).__init__(**kwargs)

    def populate(self, **kwargs):
        relations = self.prepopulate(**kwargs)
        self.prefetch_related(kwargs)
        super(Model, self).populate(**relations)

    def prepopulate(self, **kwargs):
        relations = {}
        for _, struct_name, field in self.iterate_with_name():
            if struct_name in kwargs and isinstance(
                field, (ForeignKey, ForeignKeyList, EmbeddedField, ListField)
            ):
                relations[struct_name] = kwargs.pop(struct_name)
        super(Model, self).populate(**kwargs)
        return relations

    def bind(self, parent):
        self._cache.parent = parent

    @classmethod
    def to_db_transform(cls, data):
        return data

    @classmethod
    def from_db_transform(cls, data):
        return data
    
    @classmethod
    def get_all_relateds(cls, struct):
        relateds = defaultdict(set)
        for _, struct_name, field in cls.iterate_with_name():
            if isinstance(field, ForeignKey):
                value = struct.get(struct_name, None)
                model = getattr(field, "model", None)
                if value:
                    relateds[model].add(value)
            elif isinstance(field, ForeignKeyList):
                value = pointed_getter(struct, struct_name, [])
                if isinstance(value, list):
                    model = getattr(field, "inner_model", None)
                    relateds[model].update([i for i in value if i])
            elif isinstance(field, ListField):
                values = pointed_getter(struct, struct_name, [])
                if isinstance(values, list):
                    for val in values:
                        if isinstance(val, Model):
                            main_type = val.__class__
                            val = val.to_struct()
                        else:
                            main_type = field._get_main_type()
                            if isinstance(main_type, _LazyType):
                                main_type = main_type.evaluate(cls)
                        if issubclass(main_type, Model):
                            for model, pks in main_type.get_all_relateds(val).items():
                                relateds[model].update(pks)
            elif isinstance(field, EmbeddedField):
                value = struct.get(struct_name, None)
                if isinstance(value, Model):
                    value = value.to_struct()
                if not isinstance(value, dict):
                    continue
                child_relateds = field._get_embed_type().get_all_relateds(value)
                for model, pks in child_relateds.items():
                    relateds[model].update(pks)
        return relateds

    def prefetch_related(self, struct):
        if struct.keys() and self._cache.is_root:
            relateds = self.get_all_relateds(struct)
            for model, pks in relateds.items():
                self._cache.prefetched_data[model] = (
                    {obj.pk: obj for obj in build_model_cache(model, pks)}
                    if len(pks) > 0
                    else {}
                )

    def get_prefetched_data(self):
        return self._cache.get_prefetched_data()


def build_model_cache(model, values):
    models = []
    pks = []
    for value in values:
        if isinstance(value, model):
            models.append(value)
        else:
            pks.append(value)
    models_pks = [m.pk for m in models]
    pks = [pk for pk in pks if pk not in models_pks]
    if len(pks):
        models += list(model.objects.filter(pk__in=pks))
    return models



from camomilla.models import Page
from camomilla.permissions import CamomillaBasePermissions
from camomilla.serializers import PageSerializer
from camomilla.views.base import BaseModelViewset
from camomilla.views.mixins import BulkDeleteMixin, GetUserLanguageMixin


class PageViewSet(GetUserLanguageMixin, BulkDeleteMixin, BaseModelViewset):

    queryset = Page.objects.all()
    serializer_class = PageSerializer
    permission_classes = (CamomillaBasePermissions,)
    model = Page



from rest_framework import serializers
from rest_framework.utils import model_meta

from camomilla.serializers.utils import build_standard_model_serializer
from camomilla.structured.fields import (
    EmbeddedField,
    ForeignKey,
    ForeignKeyList,
    ListField,
)
from camomilla.structured.models import Model


class StructuredJSONField(serializers.JSONField):
    def __init__(self, **kwargs):
        self.schema = kwargs.pop("schema", None)
        super().__init__(**kwargs)

    def bind(self, field_name, parent):
        if self.schema is None and isinstance(parent, serializers.ModelSerializer):
            info = model_meta.get_field_info(parent.Meta.model)
            field = info.fields[field_name]
            self.schema = field.schema
        super().bind(field_name, parent)

    def to_internal_value(self, data):
        if isinstance(data, dict):
            data = to_internal_model_value(self.schema, data)
        elif isinstance(data, list):
            data = [to_internal_model_value(self.schema, v) for v in data]
        return super().to_internal_value(data)

    def to_representation(self, instance):
        return super().to_representation(expanded_rapresentation(instance))


def expanded_model_rapresentation(schema: Model):
    stack = {}
    for _, name, field in schema.iterate_with_name():
        value = field.__get__(schema)
        if value is None:
            continue
        elif isinstance(field, ForeignKey):
            serializer = build_standard_model_serializer(field.model, depth=1)
            data = serializer(instance=value).data
        elif isinstance(field, ForeignKeyList):
            serializer = build_standard_model_serializer(field.inner_model, depth=1)
            data = serializer(instance=value, many=True).data
        elif isinstance(field, EmbeddedField):
            data = expanded_model_rapresentation(value)
        elif isinstance(field, ListField):
            data = []
            for v in value:
                if isinstance(v, Model):
                    v = expanded_model_rapresentation(v)
                data.append(v)
        else:
            data = field.to_struct(value)
        stack[name] = data
    return stack


def expanded_rapresentation(value):
    if isinstance(value, dict):
        value = {k: expanded_rapresentation(v) for k, v in value.items()}
    elif isinstance(value, list):
        value = [expanded_rapresentation(v) for v in value]
    elif isinstance(value, Model):
        value = expanded_model_rapresentation(value)
    return value


def to_internal_model_value(schema: Model, data):
    stack = {}
    for _, name, field in schema.iterate_with_name():
        value = data.get(name, None)
        if value is None:
            continue
        if isinstance(field, ForeignKey) and isinstance(value, dict):
            value = value.get(field.model._meta.pk.attname, None)
        elif isinstance(field, ForeignKeyList) and isinstance(value, list):
            attname = field.model._meta.pk.attname
            value = [v.get(attname, None) for v in value]
        elif isinstance(field, EmbeddedField):
            value = to_internal_model_value(field._get_embed_type(), value)
        elif isinstance(field, ListField):
            main_type = field._get_main_type()
            value = [to_internal_model_value(main_type, v) for v in value]
        else:
            value = field.to_struct(value)
        stack[name] = value
    return stack



from camomilla.serializers.mixins import AbstractPageMixin
from camomilla.models import Content, Page
from camomilla.serializers.base import BaseModelSerializer



class ContentSerializer(BaseModelSerializer):
    class Meta:
        model = Content
        fields = "__all__"


class PageSerializer(AbstractPageMixin, BaseModelSerializer):
    class Meta:
        model = Page
        fields = "__all__"



from functools import cached_property

from modeltranslation import settings as mt_settings
from modeltranslation.translator import NotRegistered, translator
from modeltranslation.utils import build_localized_fieldname
from rest_framework import serializers
from rest_framework.exceptions import ValidationError
from camomilla.utils.getters import pointed_getter
from camomilla.utils.translation import is_translatable


TRANS_ACCESSOR = "translations"


def plain_to_nest(data, fields, accessor=TRANS_ACCESSOR):
    trans_data = {}
    for lang in mt_settings.AVAILABLE_LANGUAGES:
        lang_data = {}
        for field in fields:
            trans_field_name = build_localized_fieldname(field, lang)
            if trans_field_name in data:
                lang_data[field] = data.pop(trans_field_name)
        if lang_data.keys():
            trans_data[lang] = lang_data
    if trans_data.keys():
        data[accessor] = trans_data
    return data


def nest_to_plain(data, fields, accessor=TRANS_ACCESSOR):
    translations = data.pop(accessor, {})
    for lang in mt_settings.AVAILABLE_LANGUAGES:
        nest_trans = translations.pop(lang, {})
        for k in fields:
            data.pop(k, None)  # this removes all trans field without lang
            if k in nest_trans:
                # this saves on the default field the default language value
                if lang == mt_settings.DEFAULT_LANGUAGE:
                    data[k] = nest_trans[k]
                data["%s_%s" % (k, lang)] = nest_trans[k]
    return data


class TranslationsMixin(serializers.ModelSerializer):
    @cached_property
    def translation_fields(self):
        try:
            return translator.get_options_for_model(self.Meta.model).get_field_names()
        except NotRegistered:
            return []

    @property
    def _writable_fields(self):
        for field in super()._writable_fields:
            if not field.field_name in self.translation_fields:
                yield field

    def to_internal_value(self, data):
        if self.translation_fields:
            nest_to_plain(data, self.translation_fields)
        return super().to_internal_value(data)

    def to_representation(self, instance):
        representation = super().to_representation(instance)
        if self.translation_fields:
            plain_to_nest(representation, self.translation_fields)
        return representation

    def run_validation(self, *args, **kwargs):
        try:
            return super().run_validation(*args, **kwargs)
        except ValidationError as ex:
            if self.translation_fields:
                plain_to_nest(ex.detail, self.translation_fields)
            raise ValidationError(detail=ex.detail)

    @property
    def is_translatable(self):
        return is_translatable(pointed_getter(self, "Meta.model"))



from django.contrib.contenttypes.models import ContentType
from django.http import Http404
from rest_framework.decorators import action
from rest_framework.response import Response

from camomilla.models import AbstractPage, Menu
from camomilla.permissions import CamomillaBasePermissions
from camomilla.serializers import ContentTypeSerializer, MenuSerializer
from camomilla.views.base import BaseModelViewset


class MenuViewSet(BaseModelViewset):
    queryset = Menu.objects.all()
    serializer_class = MenuSerializer
    permission_classes = (CamomillaBasePermissions,)
    model = Menu

    @action(detail=False, methods=["get"], url_path="page_types")
    def page_types(self, request, *args, **kwargs):
        return Response(
            ContentTypeSerializer(
                [
                    model
                    for model in ContentType.objects.all()
                    if model.model_class()
                    and issubclass(model.model_class(), AbstractPage)
                ],
                many=True,
            ).data
        )

    @action(detail=False, methods=["get"], url_path=r"page_types/(?P<content_id>\w+)")
    def page_type_instances(self, request, content_id, *args, **kwargs):
        content_type = ContentType.objects.filter(pk=content_id).first()
        if not content_type or not issubclass(content_type.model_class(), AbstractPage):
            raise Http404("No object matches the given query.")
        return Response(
            [
                {"id": obj.pk, "name": str(obj), "url_node_id": obj.url_node.pk}
                for obj in content_type.model_class().objects.exclude(url_node__isnull=True)
            ]
        )



from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.translation import gettext_lazy as _
from django.template.loader import render_to_string
from django.template import RequestContext
from django.utils.safestring import mark_safe
from camomilla import structured
from camomilla.models.page import UrlNode
from typing import Union


class MenuNodeLink(structured.Model):
    link_type = structured.CharField()
    static = structured.CharField()
    url_node = structured.ForeignKey(UrlNode)
    content_type = structured.IntegerField()
    page_id = structured.IntegerField()

    @classmethod
    def to_db_transform(cls, data):
        if data.get("link_type", None) == "RE":
            ct_id = data.get("content_type", None)
            p_id = data.get("page_id", None)
            if ct_id and p_id:
                c_type = ContentType.objects.filter(pk=ct_id).first()
                model = c_type and c_type.model_class()
                page = model and model.objects.filter(pk=p_id).first()
                if page:
                    data["url_node"] = page.url_node.pk
                else:
                    data["url_node"] = None
        return data

    def get_url(self, request=None):
        if self.link_type == "RE":
            return self.url_node and self.url_node.routerlink
        elif self.link_type == "ST":
            return self.static

    @property
    def url(self):
        return self.get_url()


class MenuNode(structured.Model):
    id = structured.CharField()
    meta = structured.DictField()
    nodes = structured.ListField(items_types=("MenuNode",))
    title = structured.CharField()
    link = structured.EmbeddedField("MenuNodeLink")

    @classmethod
    def to_db_transform(cls, data):
        link = data.pop("link", {})
        nodes = data.pop("nodes", {})
        return {
            "link": MenuNodeLink.to_db_transform(link),
            "nodes": [cls.to_db_transform(n) for n in nodes],
            **data,
        }


class Menu(models.Model):
    key = models.CharField(max_length=200, unique=True)
    available_classes = models.JSONField(default=dict)
    enabled = models.BooleanField(default=True)
    nodes = structured.StructuredJSONField(default=list, schema=MenuNode)

    class Meta:
        verbose_name = _("menu")
        verbose_name_plural = _("menus")

    def render(
        self,
        template_path: str,
        request=None,
        context: Union[dict, RequestContext] = {},
    ):
        if isinstance(context, RequestContext):
            context = context.flatten()
        context.update({"menu": self})
        return mark_safe(render_to_string(template_path, context, request))

    class defaultdict(dict):
        def __missing__(self, key):
            dict.__setitem__(self, key, Menu.objects.get_or_create(key=key)[0])
            return self[key]



from __future__ import unicode_literals
from collections import OrderedDict

from django.apps import AppConfig, apps
from django.conf import settings


class CamomillaThemeConfig(AppConfig):
    name = "camomilla.theme"

    def ready(self):
        installed_apps = getattr(settings, "INSTALLED_APPS", [])
        changed = False
        if "admin_interface" not in installed_apps:
            changed = True
            installed_apps = ["admin_interface", *installed_apps]
        if "colorfield" not in installed_apps:
            changed = True
            installed_apps = ["colorfield", *installed_apps]
        if changed:
            setattr(settings, "INSTALLED_APPS", installed_apps)
            apps.app_configs = OrderedDict()
            apps.apps_ready = apps.models_ready = apps.loading = apps.ready = False
            apps.clear_cache()
            apps.populate(settings.INSTALLED_APPS)
        setattr(settings, "X_FRAME_OPTIONS", getattr(settings, "X_FRAME_OPTIONS", ))
